/***********************************************************************
 * File: unit_test.c
 * 	(Extracted from
 *	   "$Id: unit_test.c,v 1.1 2007/10/28 04:56:34 Administrator Exp Administrator $")
 *
 * Abstracts:
 *  This is the C source file that implements 'unit test' that check the
 * 'sort_linked_list()' API.
 *
 * Usage:
 *	unit_test [-B|-Q] nodes-number
 *
 *	nodes-number : number of the testing nodes unit.
 *	-B : execute the bubble-sorting algorithm (stupid!)
 *	-Q : execute 'sort_linked_list()' API (quick sort)
 *	-L : execute 'sort_linked_list()' API (quick sort)
 *	     for looped lists.
 *	* "-B" and "-Q" aren't specified, this program execute both.
 *
 * Copyright (C) 2005-2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include "sort_linked_list.h" /* definitions of "sort_linked_list()" */

/*
 * define the node type of the bidirectional linked list.
 * this nodes are pointing to the datum via its 'key' member.
 */
typedef struct list_node_t {
	struct list_node_t	*prev;	/* previous node */
	char			*key;	/* points to datum */
#				define KEY_LEN	5
	struct list_node_t	*next;	/* next node */
} List_node_t;

/* returns the address of the next node */
static void *next(const void *list)
{	return ((List_node_t *)list)->next;
}

/* returns the address of the previous node */
static void *prev(const void *list)
{	return ((List_node_t *)list)->prev;
}

/* returns the address of the datum */
static void *get_datum_addr(const void *node)
{	return	((List_node_t *)node)->key;
}

/* returns the address of the 'key' member */
static void *get_datum_fld_addr_on_node(const void *node)
{	return	&(((List_node_t *)node)->key);
}

/* compare two nodes using 'strncmp()' */
#define CMP_NODES(N1,N2) strncmp((N1), (N2), KEY_LEN)

/* compare two nodes via pointer-to-pointer */
static int cmp_nodes_via_ptr2ptr(const void *node1, const void *node2)
{	return	CMP_NODES((*(List_node_t **)node1)->key
			, (*(List_node_t **)node2)->key
			);
}

/* compare two datum ('key') via pointer-to-pointer */
static int cmp_datum_via_ptr2ptr(const void *node1, const void *node2)
{	return	CMP_NODES(*((char **)node1), *((char **)node2));
}

/* set the next node to current node */
static void *set_next(void *src_node, const void *dest_node)
{	((List_node_t *)src_node)->next
				= (struct list_node_t *)dest_node;
	return	src_node; /* returns current node */
}

/* set the previous node to current node */
static void *set_prev(void *src_node, const void *dest_node)
{	((List_node_t *)src_node)->prev
				= (struct list_node_t *)dest_node;
	return	src_node; /* returns current node */
}

/* generate the datum ('key') */
static char *gen_key(char *keyp
		, char *key_chars
		, size_t key_chars_sz
		, ssize_t *kix
		)
{
	keyp[0]	= key_chars[kix[0]];
				kix[0] = (kix[0] + 1) % key_chars_sz;
	keyp[1]	= key_chars[kix[1]];
				kix[1] = (kix[1] + 1) % key_chars_sz;
	keyp[2]	= key_chars[kix[2]];
				kix[2] = (kix[2] + 1) % key_chars_sz;
	keyp[3]	= key_chars[kix[3]];
				kix[3] = (kix[3] + 1) % key_chars_sz;
	keyp[4] = '\0';

	return	keyp;
}

/* generate the linked list for unit-testing */
static void gen_list_and_keys(size_t	 nodes_n
				, List_node_t *list
				, char 	 *keys
				, char 	 *key_chars
				, size_t key_chars_sz
				, ssize_t *kix
				, int	 make_bidirection_list
				)
{
	char	*keyp;
	size_t	x;

	(void)set_prev(list, NULL);
	for (x = 0, keyp = keys; x < nodes_n - 1; keyp += KEY_LEN, ++x)
	{
		(void)set_next(list + x, list + (x + 1));
		(void)set_prev((list + (x + 1))
				, ((make_bidirection_list != 0)?
						list + x: NULL
					)
				);
		list[x].key	= gen_key(keyp
					, key_chars
					, key_chars_sz
					, kix
					);
	}
	(void)set_next(list + x, NULL);
	list[x].key	= gen_key(keyp
				, key_chars
				, key_chars_sz
				, kix
				);

	return;
}

/* set the index of nodes */
static void set_index(ssize_t nodes_n
			, List_node_t *list
			, List_node_t **index
			)
{
	if ((list != NULL) && index != NULL) {
		while (nodes_n-- > 0) {
			*index++	= list++;
		}
	}

	return;
}

/* print the linked list */
static void show_list(char *header
			, List_node_t *list
			, size_t nodes_n
			, List_node_t **index
			)
{
	size_t	i;

	if (nodes_n > 10)	return; /* discard the too long list */

	printf("*** dump the linked list (%p, nodes_n = %lu, "
		"index = %p) : %s...\n"
		,		(void *)list, (unsigned long)nodes_n
		,	(void *)index, header
		); (void)fflush(stdout);
	for (i = 0L
	; (i < nodes_n) && list != NULL
	; list = (List_node_t *)next(list), ++i) {
		printf("\t[%lu] index[%lu] = %p. list (%p) "
			"{ prev = %p, next = %p, "
			"key = %p [%s] }\n"
			, (unsigned long)i, (unsigned long)i
			,		(void *)((index == NULL)?
							NULL: index[i]
						)
			, (void *)list
			, (void *)(list->prev), (void *)(list->next)
			, (void *)(list->key), list->key
			); (void)fflush(stdout);
	}
	printf("\n"); (void)fflush(stdout);

	return;
}

/* print the elapsed times (usecs) to execute the sorting function */
static void show_elapsed(char *func_name
			, size_t	 nodes_n
			, struct timeval *starttv
			, struct timeval *endtv)
{
	long	delta = endtv->tv_usec - starttv->tv_usec;

	printf("*** elapsed (usec.) to execute '%s (nodes_n = %lu)' = "
		"%10ld.%06ld "
		"(start = %10ld.%06ld, "
		"end = %10ld.%06ld)\n"
		, func_name, (unsigned long)nodes_n
		, (endtv->tv_sec - starttv->tv_sec
				- ((delta < 0L)? 1L: 0L))
		, (endtv->tv_usec - starttv->tv_usec
				+ ((delta < 0L)? 1000000L: 0L))
		, starttv->tv_sec, starttv->tv_usec
		, endtv->tv_sec, endtv->tv_usec
		); (void)fflush(stdout);

	return;
}

/* bubble sorting is stupids sorting */
static void *bubble_sorter(List_node_t *list)
{
	List_node_t	*lp1, *lp2;
	char		*key;

	/* search the least datum... */
	for (lp1 = list; lp1 != NULL; lp1 = (List_node_t *)next(lp1))
	for (lp2 = (List_node_t *)next(lp1)
	; lp2 != NULL
	; lp2 = (List_node_t *)next(lp2)
	) {
		if (CMP_NODES(lp1->key, lp2->key) > 0) {
			/* ... swap those */
			key	= lp1->key;
			lp1->key	= lp2->key;
			lp2->key	= key;
		}
	}

	return	list; /* returns the first node of the original list */
}

/* Unit-test program entry */
int main(int ac, char **av)
{
#	define KEY_CHARS	"aZbYcXdWeVfUgThSiRjQkPlOmN"	\
				"nMoLpKqJrIsHtGuFvEwDxCyBzA"
	List_node_t	*list, *result, **index;
	char		*keys = NULL, *NULLp = NULL;
	size_t		list_allocate_sz = 0;
	ssize_t		max_nodes_n
			, nodes_n = 0
			, key_chars_sz = strlen(KEY_CHARS)
			, kix[KEY_LEN - 1] = { 0L, 1L, 2L, 4L };
	long		sorter_options = 0, i = 1, index_needs;
	struct timeval	starttv, endtv;

	/* test command option is nothing? */
	if (ac == 1) {
USAGE:		/* print usage message */
		fprintf(stderr
			, "Usage: %s [-B|Q] node-number\n"
			,	av[0]
			); (void)fflush(stderr);
		exit(1); /* usage error */
	}

	/* parse the command options */
	if (strcmp(av[i], "-B") == 0) {
		/* execute bubble sorting */
#		define BUBBLE_SORTING 1
		sorter_options |= BUBBLE_SORTING;
		++i;
	}
	else if (strcmp(av[i], "-Q") == 0) {
		/* execute quick sorting */
#		define QUICK_SORTING 2
		sorter_options |= QUICK_SORTING;
		++i;
	}
	else if (strcmp(av[i], "-L") == 0) {
		/* execute quick sorting on the looped list */
#		define QUICK_SORTING_LOOPED 4
		sorter_options |= QUICK_SORTING_LOOPED;
		++i;
	}
	if (sorter_options == 0) {
		/* all sorting methods */
		sorter_options = (BUBBLE_SORTING
				| QUICK_SORTING
				| QUICK_SORTING_LOOPED
				);
	}

	/* convert the nodes-number option to 'unsigned long' */
	if ((i != ac - 1)
	|| (nodes_n = (size_t)strtoul(av[i], NULL, 10)) == 0L)
		goto USAGE; /* usage error */

	/* allocate the test data area (linked list and key strings) */
	#define MAX_EXTEND_NODES_N	8
	#define STRIDE_OF_EXTEND_NODE	2
	list_allocate_sz = MAX_EXTEND_NODES_N
				* ((sizeof(*list) + KEY_LEN) * nodes_n);
	if ((list = (List_node_t *)malloc(list_allocate_sz)) == NULL) {
		fprintf(stderr
			, "***FATAL:L%d: failed to malloc() "
			  "(list = %p. keys = %p).\n"
			,	__LINE__
			,	(void *)list,		(void *)keys
			); (void)fflush(stderr);
		exit(2); /* insufficient memories */
	}
	memset(list, (int)'\0', list_allocate_sz);

	/* print the starting message */
	printf(">>> Unit Test for 'sort_linked_list()' starts...\n"
		); (void)fflush(stdout);

	/* iterate: extend the numbers of nodes (multiplies 2) */
	for (max_nodes_n = MAX_EXTEND_NODES_N * nodes_n
	; nodes_n <= max_nodes_n
	; nodes_n *= STRIDE_OF_EXTEND_NODE) {
		/* set the datum area address to 'keys' */
		keys	= ((char *)list) + (sizeof(*list) * nodes_n);

		/* test need to execute bubble-sorting? */
		if ((sorter_options & BUBBLE_SORTING) != 0) {
			/* generate the linked list and datums */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 1 /* make bidirection list */
					);

			/* print the initialized linked list */
			show_list("bubble_sorter (before)"
				, list
				, nodes_n
				, NULL
				);

			/* execute bubble sorting */
			gettimeofday((struct timeval *)memset(
						  &starttv
						, (int)'\0'
						, sizeof(starttv)
						)
					, NULL
					);
			printf("bubble_sorter(%p) "
				"returns %p.\n"
				,	(void *)list
				,	(void *)bubble_sorter(list)
				); (void)fflush(stdout);
			gettimeofday((struct timeval *)memset(
						  &endtv
						, (int)'\0'
						, sizeof(endtv)
						)
					, NULL
					);

			/* print sorted linked list */
			show_list("bubble_sorter (after)"
				, list
				, nodes_n
				, NULL
				);

			/* print elapsed times */
			show_elapsed("bubble_sorter()"
					, nodes_n
					, &starttv
					, &endtv
					);
		}

		/* test need to execute quick sorting using
		 * 'sort_linked_list()'
		 */
		if ((sorter_options & QUICK_SORTING) != 0) {
		for (index_needs = 0; index_needs < 2; ++index_needs) {
			{ /* sort datum in the bidirection list */
			/* generate the bidirection linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 1 /* make bidirection list */
					);

			/* print the initialized linked list */
			show_list("sort_datum_in_linked_list("
					"bidirection) (before)"
				, list
				, nodes_n
				, (index = NULL)
				);

			/* execute quick sorting (sorting datum) */
			gettimeofday((struct timeval *)memset(
						  &starttv
						, (int)'\0'
						, sizeof(starttv)
						)
					, NULL
					);
			printf("sort_datum_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, (void *)prev
				, (void *)get_datum_addr
				, (void *)get_datum_fld_addr_on_node
				, (void *)cmp_datum_via_ptr2ptr
				, ((index_needs == 0)?
					  (void *)(&index): NULL)
				, (void *)(result = (List_node_t *)
					sort_datum_in_linked_list(
					  list
					, nodes_n
					, next
					, prev
					, get_datum_addr
					, get_datum_fld_addr_on_node
					, cmp_datum_via_ptr2ptr
					, set_prev
					, ((index_needs == 0)?
						  (void ***)(&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday((struct timeval *)memset(
						  &endtv
						, (int)'\0'
						, sizeof(endtv)
						)
					, NULL
					);

			/* print sorted linked list */
			show_list("sort_datum_in_linked_list("
					"bidirection) (after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			free(index); index = NULL;

			/* print elapsed times */
			show_elapsed("sort_datum_in_linked_list("
						"bidirection)"
					, nodes_n
					, &starttv
					, &endtv
					);

			/* test the result means illegal? */
			if (result != list) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_datum_in_linked_list("
				  			"bidirection) "
				  "(result = %p, list = %p).\n"
				,	__LINE__
				,	(void *)result,	(void *)list
				); (void)fflush(stderr);
				exit(4); /* FATAL */
			}

			} /* sort datum in the bidirection list */

			{ /* sort nodes in the unidirectional list */
			
			/* generate the indirection linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 0 /* make indirection list */
					);

			/* print the initialized linked list */
			show_list("sort_nodes_in_linked_list(unidirect)"
					" (before)"
				, list
				, nodes_n
				, (index = NULL)
				);

			/* execute quick sorting (sort nodes) */
			gettimeofday(&starttv, NULL);
			printf("sort_nodes_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, NULLp
				, (void *)cmp_nodes_via_ptr2ptr
				, (void *)set_next
				, NULLp
				, ((index_needs == 0)?
					  (void *)(&index): NULL)
				, (void *)(result = (List_node_t *)
					sort_nodes_in_linked_list(
					  list
					, nodes_n
					, next
					, NULL
					, cmp_nodes_via_ptr2ptr
					, set_next
					, NULL /* unidirectional list */
					, ((index_needs == 0)?
						  (void ***)(&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday(&endtv, NULL);

			/* print sorted linked list */
			show_list("sort_nodes_in_linked_list(unidirect)"
					" (after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			free(index); index = NULL;

			/* print elapsed times */
			show_elapsed("sort_nodes_in_linked_list("
						"unidirect)"
					, nodes_n
					, &starttv
					, &endtv
					);

			/* test the result means illegal? */
			if (result == NULL) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_linked_list(unidirect) "
				  "(result = %p).\n"
				,	__LINE__
				,		(void *)result
				); (void)fflush(stderr);
				exit(3); /* FATAL */
			}

			} /* sort nodes in the unidirectional list */

			{ /* sort nodes in the bidirection list */
			/* generate the bidirection linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 1 /* make bidirection list */
					);

			/* print the initialized linked list */
			show_list("sort_nodes_in_linked_list(bidirect) "
					"(before)"
				, list
				, nodes_n
				, (index = NULL)
				);

			/* execute quick sorting (sort nodes) */
			gettimeofday(&starttv, NULL);
			printf("sort_nodes_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, (void *)prev
				, (void *)cmp_nodes_via_ptr2ptr
				, (void *)set_next
				, (void *)set_prev
				, ((index_needs == 0)?
					  (void *)(&index): NULL)
				, (void *)(result = (List_node_t *)
					sort_nodes_in_linked_list(
					  list
					, nodes_n
					, next
					, prev
					, cmp_nodes_via_ptr2ptr
					, set_next
					, set_prev /* bidirection */
					, ((index_needs == 0)?
						  (void ***)(&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday(&endtv, NULL);

			/* print sorted linked list */
			show_list("sort_nodes_in_linked_list(bidirect) "
					"(after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			free(index); index = NULL;

			/* print elapsed times */
			show_elapsed("sort_nodes_in_linked_list("
						"bidirect)"
					, nodes_n
					, &starttv
					, &endtv
					);

			/* test the result means illegal? */
			if (result == NULL) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_linked_list(bidirect) "
				  "(result = %p).\n"
				,	__LINE__
				,		(void *)result
				); (void)fflush(stderr);
				exit(3); /* FATAL */
			}

			} /* sort nodes in the bidirection list */
		}
		}

		/* test need to execute quick sorting looped lists,
		 * using 'sort_linked_list()'
		 */
		if ((sorter_options & QUICK_SORTING_LOOPED) != 0) {
		for (index_needs = 0, index = NULL
		; index_needs < 4
		; ++index_needs) {
			{ /* sort datum in the bidirection list */
			/* generate the bidirection linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 1 /* make bidirection list */
					);

			/* set the index array */
			set_index(nodes_n, list, index);

			/* make the list to be looped */
			if (index_needs < 2) { /* ring buffer */
				(void)set_prev(list
						, list + (nodes_n - 1)
						);
				(void)set_next((list + (nodes_n - 1))
						, list
						);
			}
			else { /* looped linear graph */
				(void)set_next(list + (nodes_n - 1)
						, list + (nodes_n / 2)
						);
			}

			/* print the initialized linked list */
			show_list("sort_datum_in_linked_list("
					"bidirection looped) (before)"
				, list
				, nodes_n
				, index
				);

			/* execute quick sorting (sorting datum) */
			gettimeofday((struct timeval *)memset(
						  &starttv
						, (int)'\0'
						, sizeof(starttv)
						)
					, NULL
					);
			printf("sort_datum_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, (void *)prev
				, (void *)get_datum_addr
				, (void *)get_datum_fld_addr_on_node
				, (void *)cmp_datum_via_ptr2ptr
				, ((index_needs % 2 == 0)?
					  (void *)(&index): NULL)
				, (void *)(result = (List_node_t *)
					sort_datum_in_linked_list(
					  list
					, nodes_n
					, next
					, prev
					, get_datum_addr
					, get_datum_fld_addr_on_node
					, cmp_datum_via_ptr2ptr
					, set_prev
					, ((index_needs % 2 == 0)?
						  (void ***)(&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday((struct timeval *)memset(
						  &endtv
						, (int)'\0'
						, sizeof(endtv)
						)
					, NULL
					);

			/* print sorted linked list */
			show_list("sort_datum_in_linked_list("
					"bidirection looped) (after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			if (index_needs % 2 == 1) {
				free(index); index = NULL;
			}

			/* print elapsed times */
			show_elapsed((char *)((index_needs < 2)?
					  "sort_datum_in_linked_list("
					  "bidirection ring-buffer)"
					: "sort_datum_in_linked_list("
					  "bidirection looped-graph)"
					)
				, nodes_n
				, &starttv
				, &endtv
				);

			/* test the result means illegal? */
			if (result != list) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_datum_in_linked_list("
						"bidirection %s) "
				  "(result = %p, list = %p).\n"
				,	__LINE__
				,	((index_needs < 2)?
						  "ring-buffer"
						: "looped-graph"
						)
				,	(void *)result,	(void *)list
				); (void)fflush(stderr);
				exit(4); /* FATAL */
			}

			} /* sort datum in the bidirection list */

			{ /* sort nodes in the unidirectional list */
			
			/* generate the unidirectional linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 0 /* make unidirectional */
					);

			/* set the index array */
			set_index(nodes_n, list, index);

			/* make the list to be looped */
			if (index_needs < 2) { /* ring buffer */
				(void)set_prev(list
						, list + (nodes_n - 1)
						);
				(void)set_next((list + (nodes_n - 1))
						, list
						);
			}
			else { /* looped linear graph */
				(void)set_next(list + (nodes_n - 1)
						, list + (nodes_n / 2)
						);
			}

			/* print the initialized linked list */
			show_list("sort_nodes_in_linked_list(unidirect "
					"looped) (before)"
				, list
				, nodes_n
				, index
				);

			/* execute quick sorting (sort nodes) */
			gettimeofday(&starttv, NULL);
			printf("sort_nodes_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, NULLp
				, (void *)cmp_nodes_via_ptr2ptr
				, (void *)set_next
				, NULLp
				, ((index_needs % 2 == 0)?
					(void *)(&index): NULLp)
				, (void *)(result = (List_node_t *)
					sort_nodes_in_linked_list(
					  list
					, nodes_n
					, next
					, (void*(*)(const void *))NULL
					, cmp_nodes_via_ptr2ptr
					, set_next
					, (void*(*)(void*, const void*))
					  NULL /* unidirectional list */
					, (void ***)(
						(index_needs % 2 == 0)?
						  (&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday(&endtv, NULL);

			/* print sorted linked list */
			show_list("sort_nodes_in_linked_list(unidirect "
					"looped) (after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			if (index_needs % 2 == 1) {
				free(index); index = NULL;
			}

			/* print elapsed times */
			show_elapsed((char *)((index_needs < 2)?
					  "sort_nodes_in_linked_list("
					  "unidirect ring-buffer)"
					: "sort_nodes_in_linked_list("
					  "unidirect looped-graph)"
					)
				, nodes_n
				, &starttv
				, &endtv
				);

			/* test the result means illegal? */
			if (result == NULL) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_linked_list(unidirect %s) "
				  "(result = %p).\n"
				,	__LINE__
				,	((index_needs < 2)?
						  "ring-buffer"
						: "looped-graph"
						)
				,		(void *)result
				); (void)fflush(stderr);
				exit(3); /* FATAL */
			}

			} /* sort nodes in the unidirectional list */

			{ /* sort nodes in the bidirection list */
			/* generate the bidirection linked list
			 * and datums
			 */
			gen_list_and_keys(nodes_n
					, list
					, keys
					, KEY_CHARS
					, key_chars_sz
					, kix
					, 1 /* make bidirection list */
					);

			/* set the index array */
			set_index(nodes_n, list, index);

			/* make the list to be looped */
			if (index_needs < 2) { /* ring buffer */
				(void)set_prev(list
						, list + (nodes_n - 1)
						);
				(void)set_next((list + (nodes_n - 1))
						, list
						);
			}
			else { /* looped linear graph */
				(void)set_next(list + (nodes_n - 1)
						, list + (nodes_n / 2)
						);
			}

			/* print the initialized linked list */
			show_list("sort_nodes_in_linked_list(bidirect "
					"looped) (before)"
				, list
				, nodes_n
				, index
				);

			/* execute quick sorting (sort nodes) */
			gettimeofday(&starttv, NULL);
			printf("sort_nodes_in_linked_list"
				"(%p, %ld, %p, %p, %p, %p, %p, %p) "
				"returns %p.\n"
				, (void *)list, (long)nodes_n
				, (void *)next, (void *)prev
				, (void *)cmp_nodes_via_ptr2ptr
				, (void *)set_next
				, (void *)set_prev
				, ((index_needs % 2 == 0)?
					(void *)(&index): NULL)
				, (void *)(result = (List_node_t *)
					sort_nodes_in_linked_list(
					  list
					, nodes_n
					, next
					, prev
					, cmp_nodes_via_ptr2ptr
					, set_next
					, set_prev /* bidirection */
					, ((index_needs % 2 == 0)?
						  (void ***)(&index)
						: NULL
						)
					))
				); (void)fflush(stdout);
			gettimeofday(&endtv, NULL);

			/* print sorted linked list */
			show_list("sort_nodes_in_linked_list(bidirect "
					"looped) (after)"
				, result
				, nodes_n
				, index
				);

			/* free the allocated area for 'index' */
			if (index_needs % 2 == 1) {
				free(index); index = NULL;
			}

			/* print elapsed times */
			show_elapsed((char *)((index_needs < 2)?
					  "sort_nodes_in_linked_list("
					  "bidirect ring-buffer)"
					: "sort_nodes_in_linked_list("
					  "bidirect looped-graph)"
					)
				, nodes_n
				, &starttv
				, &endtv
				);

			/* test the result means illegal? */
			if (result == NULL) {
				fprintf(stderr
				, "***FATAL:L%d: failed to "
				  "sort_linked_list(bidirect %s) "
				  "(result = %p).\n"
				,	__LINE__
				,	((index_needs < 2)?
						  "ring-buffer"
						: "looped-graph"
						)
				,		(void *)result
				); (void)fflush(stderr);
				exit(3); /* FATAL */
			}

			} /* sort nodes in the bidirection list */
		}
		}
	}

	/* release the allocated test data area */
	free(list);

	/* print ending message */
	printf("<<< Unit Test for 'sort_linked_list()' finished!\n"
		); (void)fflush(stdout);

	exit(0); /* Normal exit */
}
